' C3 Flash memory object
' Code taken from Andre Lamothe's C3 Test Suite
CON

  CLOCKS_PER_MICROSECOND = 5*16     ' simple xin*pll / 1_000_000                     
  CLOCKS_PER_MILLISECOND = 5000*16  ' simple xin*pll / 1_000


  ' SPI pins
  SPI_SEL_CLK            = 8  ' used to clock the SPI counter that feeds 4-bit value to SPI selection decoder, clocks on rising edge
  SPI_SEL_CLR            = 25 ' used to clear the SPI counter back to 0000b (NULL SPI device), active LOW

  SPI_MOSI               = 9  ' SPI master out serial in to slave
  SPI_MISO               = 10 ' SPI master in serial out from slave, 10 for good C3, change to 7 for BROKEN C3 with YELLOW jumper.
  SPI_SCK                = 11 ' SPI clock from master to all slaves

  ' SPI channels defines 
  SPI_CHAN_NULL          = 0 ' NULL channel, disables all on/off board devices.
  SPI_CHAN_SRAM0         = 1 ' 32K SRAM Bank 0.
  SPI_CHAN_SRAM1         = 2 ' 32K SRAM Bank 1.
  SPI_CHAN_FLASH         = 3 ' 1MB FLASH Memory.
  SPI_CHAN_AD            = 4 ' MCP3202 2-Channel 12-bit A/D.
  SPI_CHAN_SD            = 5 ' Micro SD Card.
'//////////////////////////////////////////////////////////////////////////////
' FLASH MEMORY DEFINES ////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

' defines for the proper FLASH installed on C3 4MB, 8MB, 16MB
' 4Mb (512K Bytes)
AT26F004        = 4  ' numeric codes for flash version
AT25DF041A      = 4

' 8Mb (1M Bytes)
AT26DF081A      = 8
AT25DF081A      = 8

' 16 Mb (2M Bytes)
AT25DF161       = 16

' set type of FLASH ROM currently installed in C3 (8MBit) for conditional compilation of code -- not supported in SPIN, so manually done below :( 
FLASH_ROM             = AT26DF081A 
NUM_FLASH_ROM_SECTORS = 19 ' 11 for 4MB, 19 for 8MB, must be set manually due to lack of conditional compilation


' Atmel 2F004/8xxx series basic commands
WRITE_ENABLE     = $06
WRITE_DISABLE    = $04
READ_STATUS      = $05
WRITE_STATUS     = $01
READ_DATA        = $03

SEQ_PROGRAM      = $AF  
BYTE_PROGRAM     = $02 

BLOCK_ERASE4     = $20
BLOCK_ERASE32    = $52
BLOCK_ERASE64    = $D8

SECTOR_ERASE     = $20    

CHIP_ERASE       = $60 
JEDEC_ID         = $9F

PROTECT_SECTOR   = $36 
UNPROTECT_SECTOR = $39 

' custom commands we make up, not part of chip spec, just used here, start at code 0xF0
UNPROTECT_CHIP   = $F0

DAT

'  conditional compilation for FLASH memory variants not supported, so just comment out either declaration

' #if (FLASH_ROM == AT26F004)

'// we could use PGM memory for these and or do it algorithmically, but this is cleaner for now
'// 11 sectors [64K, 64K, 64K, 64K, 64K, 64K, 64K, 32K, 8K, 8k, 16K ]
'//            |<------- (7) 64K sectors ------->|  1      2     1 

{
    ' 11 sectors in 4MB FLASH
    sector_addresses long $00000   ' 64K sector 0
                     long $10000   ' 64K sector 1
                     long $20000   ' 64K sector 2
                     long $30000   ' 64K sector 3
                     long $40000   ' 64K sector 4
                     long $50000   ' 64K sector 5
                     long $60000   ' 64K sector 6
                     long $70000   ' 32K sector 7
                     long $78000   ' 8K sector 8  
                     long $7A000   ' 8K sector 9
                     long $7D000   ' 8K sector 10
}

' #if (FLASH_ROM == AT26DF081A)

' we could use PGM memory for these and or do it algorithmically, but this is cleaner for now
' 19 sectors [64K, 64K, 64K, ..., 64K, 64K, 64K, 64K, 16K, 8K, 8k, 32K ]
'            |<----------- (15) 64K sectors -------->| 1     2     1 

    ' 19 sectors in 8MB FLASH
    sector_addresses long $00000   ' 64K sector 0
                     long $10000   ' 64K sector 1
                     long $20000   ' 64K sector 2
                     long $30000   ' 64K sector 3
                     long $40000   ' 64K sector 4
                     long $50000   ' 64K sector 5
                     long $60000   ' 64K sector 6
                     long $70000   ' 64K sector 7 
                     long $80000   ' 64K sector 8
                     long $90000   ' 64K sector 9
                     long $A0000   ' 64K sector 10
                     long $B0000   ' 64K sector 11
                     long $C0000   ' 64K sector 12
                     long $D0000   ' 64K sector 13
                     long $E0000   ' 64K sector 14
                          
                     long $F0000   ' 16K sector 15
                     long $F4000   '  8K sector 16
                     long $F6000   '  8K sector 17
                          
                     long $F8000    ' 32K sector 18




PUB Start

  ' initialize SPI bus system
  SPI_Init ' open the flash memory up (unprotect it)
  Flash_Open
' Flash_Erase_Chip(1)
' /////////////////////////////////////////////////////////////////////////////
' FLASH MEMORY API (ported from my C/C++ libraries note syntax leftovers!)
' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Read_Status | _spi_data
{{
Reads the FLASH memory status register

PARMS: None.
RETURNS: status register value
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, READ_STATUS, $FF )

  ' now read data returned
  _spi_data := SPI_Write_Read( 8, $FF, $FF )
   
  '// disable SPI interface
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' return results
  return ( _spi_data )   

' // end Flash_Read_Status

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Write_Status ( status ) | _spi_data
{{
Writes the FLASH memory status register

PARMS:

status - value to update the status register with

RETURNS: None.
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_STATUS, $FF )
  _spi_data := SPI_Write_Read( 8, status, $FF )  
   
  '// disable SPI interface
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' return success for now
  return ( 1 )   

' // end Flash_Write_Status

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Open | _index, _spi_data, _flash_start_addr
{{

This functions "opens" the FLASH memory simply by unprotecting all sectors starting at address 0

PARMS: None.
RETURNS: None.

}}




  ' use global unprotect to unprotect entire chip at once, write "0" to bits 2,3,4,5
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
  SPI_Select_Channel( 0 )

  Flash_Write_Status ( Flash_Read_Status & %11_0000_11 )
  Delay_MS( 10 )   
  ' // return success always
  return( 1 )
  

{  
  ' this code is used to iterate thru all sectors manually and have more control
  ' iterate thru all sectors and unprotect them
  repeat _index from 0 to NUM_FLASH_ROM_SECTORS-1

    '// STEP 1: enable writing to chip  ///////////////////////////////////////
    '// enable SPI interface

    '// set CS to SPI select channel 3 (FLASH)
    SPI_Select_Channel( SPI_CHAN_FLASH )
    _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )

    '// disable SPI interface
    '// set CS to SPI select channel 0 (null)
    SPI_Select_Channel( 0 )

    Delay_MS( 10 )

    '// unprotect sector  ///////////////////////////////////

    '// enable SPI interface

    '// set CS to SPI select channel 3 (FLASH)
    SPI_Select_Channel( SPI_CHAN_FLASH )
    _spi_data := SPI_Write_Read( 8, UNPROTECT_SECTOR, $FF )

    '// get sector address
    _flash_start_addr := sector_addresses[ _index ]

    '// write sector address, there are only a handful of sectors since each is 4-64K
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 16, $FF )
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 8, $FF )
    _spi_data := SPI_Write_Read( 8, _flash_start_addr >> 0, $FF )

    '// disable SPI interface
    '// set CS to SPI select channel 0 (null)
    SPI_Select_Channel( 0 )

    ' delay 10
    Delay_MS( 10 ) 

    ' // end for index

  ' // return success always
  return( 1 )
}


' // end Flash_Open

' /////////////////////////////////////////////////////////////////////////////

PUB Flash_Close | _index, _spi_data, _flash_start_addr      
{{
This function "closes" the FLASH by protecting all of the sectors on the flash chip.

PARMS: None.
RETURNS: None.

}}

  '// protecting the sectors seems to have a side effect, we are not considering???
  '// we need to read the data sheet more, for now, just return, and don't protect 
  '// all the sectors, this is overkill anyway for now ... 
  return(1)


' // end Flash_Close
PUB Flash_Write( flash_start_addr, buffer_ptr, num_bytes) | _index, _spi_data
{{
This function writes num_bytes bytes to the FLASH memory at flash_start_addr from the sent buffer_ptr.

It uses the byte/page write function. Although, there is a "sequential" byte write function available
on the AT26DF081A the new rev of the Atel chips (AT25xxxx series) coming out in 2011 have replaced the sequential write with
another function :(, thus we simply have to check for page boundaries and re-issue another page write
every 256 bytes.

PARMS:

flash_start_addr - starting address in the FLASH memory to write the data at.
buffer_ptr       - local byte pointer to Prop memory holding data to write to FLASH memory.
num_bytes        - number of bytes to write to FLASH from source buffer_ptr. 

}}

  '// STEP 1: enable writing to the chip again
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )
   
  '// STEP 2: write buffer to address in flash //////////////////
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
   
  '// initiate byte/page program command
  _spi_data := SPI_Write_Read( 8, BYTE_PROGRAM, $FF )
   
  '// write starting address byte by byte, MSB first...
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 16, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 8, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 0, $FF )

  '// write first byte of 
  _spi_data := SPI_Write_Read( 8, byte [buffer_ptr][ 0 ], $FF )

  _index := 1   

  '// write remaining bytes
  repeat until (_index => num_bytes )

    ' now test if we are on a page boundary, if so finish page, and start new page
    if ( ((flash_start_addr + _index) // 256) == 0)
      '// set CS to SPI select channel 0 (null)
      SPI_Select_Channel( 0 )

      ' replace this with polling...
      Delay_MS(10) 

      '// set CS to SPI select channel 3 (FLASH)
      SPI_Select_Channel( SPI_CHAN_FLASH )
      _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
      SPI_Select_Channel( 0 )

      SPI_Select_Channel( SPI_CHAN_FLASH )
               
      '// re-issue byte/page program command
      _spi_data := SPI_Write_Read( 8,BYTE_PROGRAM,$FF )

      '// write starting address byte by byte, MSB first...
      _spi_data := SPI_Write_Read( 8, (flash_start_addr + _index) >> 16, $FF )
      _spi_data := SPI_Write_Read( 8, (flash_start_addr + _index) >> 8, $FF )
      _spi_data := SPI_Write_Read( 8, (flash_start_addr + _index) >> 0, $FF )

      ' end if 

    '// write the next byte of data
    _spi_data := SPI_Write_Read( 8, byte [buffer_ptr][ _index ], $FF )

    _index++
    
    ' // end for index    

  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' write final bytes left in internal buffer, replace this with polling...
  Delay_MS(10) 
     
  '// disable SPI interface
   
  '// return success
  return(1)
   
' // end Flash_Write

' ///////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

PUB Flash_Read( flash_start_addr, buffer_ptr, num_bytes) | _index, _spi_data
{{
This function reads num_bytes from the FLASH memory starting at memory location flash_start_addr and stores the data in
buffer_ptr.

PARMS:

flash_start_addr - starting address in FLASH memory to read data from.
buffer_ptr       - local Prop byte pointer to store the read data at.
num_bytes        - the number of bytes to read from the FLASH memory. 

RETURNS: None.
}}

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )

  '// initiate read command
  _spi_data := SPI_Write_Read( 8, READ_DATA, $FF )

  '// write starting address byte by byte, MSB first...
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 16, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 8, $FF )
  _spi_data := SPI_Write_Read( 8, flash_start_addr >> 0, $FF )
  
  '// now read data 
  repeat _index from  0 to num_bytes-1
    byte [buffer_ptr][_index] := SPI_Write_Read( 8, $00, $FF )

    ' // end for index

  '// disable SPI interface

  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  '// return success 
  return(1)

' // end Flash_Read

' /////////////////////////////////////////////////////////////////////////////

'//////////////////////////////////////////////////////////////////////////////
' FLASH MACROS ////////////////////////////////////////////////////////////////
'//////////////////////////////////////////////////////////////////////////////

PUB FLASH_ADDR_TO_BLOCK( n )
  n := (n >> 12)

' /////////////////////////////////////////////////////////////////////////////
  
PUB FLASH_BLOCK_TO_ADDR( n )
  n := (n << 12)

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_MS( time )
{{
Delays "time" milliseconds and returns.
}}

  waitcnt (cnt + time*CLOCKS_PER_MILLISECOND)

' // end Delay_MS

' /////////////////////////////////////////////////////////////////////////////

PUB Delay_US( time )
{{
Delays "time" microseconds and returns.

}}

  waitcnt (cnt + time*CLOCKS_PER_MICROSECOND)


' // end Delay_US
' /////////////////////////////////////////////////////////////////////////////
PUB SPI_Write_Read( num_bits, data_out, bit_mask) | data_in
{{
This function writes and reads SPI data a bit at a time (SPI is a circular buffer protocal), the data is in MSB to LSB format
and up to 32-bits can be transmitted and received, the final result is bit masked by bit_mask

PARMS:

num_bits : number of bits to transmit from data_out
data_out : source of data to transmit
bit_mask : final result of SPI transmission is masked with this to grab the relevant least significant bits

RETURNS: data retrieved from SPI transmission

}}
  ' clear result
  data_in := 0

  ' now read the bits in/out
  repeat num_bits
    ' drop clock
    OUTA[SPI_SCK] := 0

    ' place next bit on MOSI
    OUTA[SPI_MOSI] := ((data_out >> (num_bits-- - 1)) & $01)
 
    ' now read next bit from MISO
    data_in := (data_in << 1) + INA[SPI_MISO]
      
    ' raise clock
    OUTA[SPI_SCK] := 1

  ' set clock and MOSI to LOW on exit
  OUTA[SPI_MOSI]  := 0
  OUTA[SPI_SCK]   := 0

  ' at this point, the data has been written and read, return result
  return ( data_in & bit_mask )            
    
' end SPI_Write_Read

' /////////////////////////////////////////////////////////////////////////////

PUB SPI_Select_Channel( channel )
{{
This function sets the active SPI channel chip select on the SPI mux, this is accomplished by first resetting the
SPI counter that feeds the SPI select decoder, then upcounting the requested number of channels.

PARMS:

channel : channel 0 to 7 to enable where the channels are defined as follows
          0 - NULL channel, disables all on/off board devices.
          1 - 32K SRAM Bank 0.
          2 - 32K SRAM Bank 1.
          3 - 1MB FLASH Memory.
          4 - MCP3202 2-Channel 12-bit A/D.
          5 - Micro SD Card.
          6 - Header Interface SPI6.
          7 - Header Interface SPI7.

RETURNS: nothing.

}}

  ' requesting channel 0? If so, easy reset
  if (channel == 0)
    ' clear the 161
    OUTA[SPI_SEL_CLR] := 0 ' CLR counter
    OUTA[SPI_SEL_CLR] := 1 ' allow counting
    return

  ' else non-null channel, count up to channel...
  
  ' first reset the SPI channel counter
  ' clear the 161
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' now increment to requested channel
  ' clock the 161
   OUTA[SPI_SEL_CLK] := 0

   repeat channel
     OUTA[SPI_SEL_CLK] := 1
     OUTA[SPI_SEL_CLK] := 0        

' end SPI_Select_Channel
PUB SPI_Init
{{
This function initializes the SPI IO's, counter, and mux, selects channel 0 and returns.

PARMS: none.
RETURNS: nothing.
}}
  ' set SPI mux counter IOs up
  DIRA[SPI_SEL_CLK] := 1 ' set to output
  OUTA[SPI_SEL_CLK] := 0

  DIRA[SPI_SEL_CLR] := 1 ' set to output
  OUTA[SPI_SEL_CLR] := 0 ' CLR counter
  OUTA[SPI_SEL_CLR] := 1 ' allow counting

  ' set up SPI lines
  OUTA[SPI_MOSI]     := 0 ' set to LOW
  OUTA[SPI_SCK]      := 0 ' set to LOW
 
  DIRA[SPI_MOSI]     := 1 ' set to output
  DIRA[SPI_MISO]     := 0 ' set to input
  DIRA[SPI_SCK]      := 1 ' set to output

  ' set SPI to NULL channel 0
  SPI_Select_Channel( 0 )

' end SPI_Init

PUB Flash_Erase_Chip( wait ) | _spi_data
{{
This function erases the entire chip. The erase command takes a considerable amount
of time, thus this function issues the command and either returns immediately or waits.
To determine if the erase is complete the function polls the status register bit 0
which is the RDY/BSY bit, 0=device is idle, 1=device is busy

PARMS: wait - boolean, 1=wait for erasure to complete, 0-return immediately.

RETURNS: None.

}}
  '// enable SPI interface
   
  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, WRITE_ENABLE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  '// set CS to SPI select channel 3 (FLASH)
  SPI_Select_Channel( SPI_CHAN_FLASH )
  _spi_data := SPI_Write_Read( 8, CHIP_ERASE, $FF )
   
  '// disable SPI interface
   
  '// set CS to SPI select channel 0 (null)
  SPI_Select_Channel( 0 )

  ' test if user requested wait for completion?
  if (wait)
    repeat
      ' erase operation still ongoing?
      if ((Flash_Read_Status & $01)==0)
        return (1)
     
  '// return success for now
  return(1)
